use std::{borrow::Borrow, ops::Deref, str};

#[cfg(feature = "serialize")]
use serde::Serialize;

use crate::ValidationResult;

#[derive(thiserror::Error, Debug, Clone, Copy, PartialEq, Eq)]
#[cfg_attr(feature = "serialize", derive(Serialize))]
pub enum PhoneNumberError {
    #[error("Phone number must be exactly 11 characters")]
    InvalidLength,
    #[error("The first character must be a 0")]
    InvalidFirstCharacter,
    #[error("The byte ({1}) at index ({0}) is not an ascii digit")]
    InvalidCharacter(usize, u8),
}

#[derive(Debug, Clone, Copy, PartialEq, Eq, Hash)]
#[repr(transparent)]
pub struct PhoneNumber(pub(crate) [u8; 11]);

impl PhoneNumber {
    /// Phone numbers must be 11 digits and start with 0
    pub fn new(value: impl AsRef<[u8]>) -> ValidationResult<Self> {
        // Since this function is generic and will have multiple versions stamped during
        // monomorphisation, it's better to keep the validation code in the non-generic new.
        //
        // Also, to_owned for PhoneNumberRef doesn't actually allocate, so no extra cost here :)
        Ok(PhoneNumberRef::new(value.as_ref())?.to_owned())
    }

    pub fn as_str(&self) -> &str {
        // Safety:
        // The validation in the constructor makes sure that all the bytes are ascii.
        unsafe { str::from_utf8_unchecked(&self.0) }
    }
}

impl Borrow<PhoneNumberRef> for PhoneNumber {
    fn borrow(&self) -> &PhoneNumberRef {
        let value = self.0.as_slice();

        // Safety:
        // PhoneNumberRef is a wrapper of [u8], therefore converting a &[u8] to a &PhoneNumberRef is
        // safe.
        unsafe { &*(value as *const _ as *const PhoneNumberRef) }
    }
}

impl AsRef<PhoneNumberRef> for PhoneNumber {
    fn as_ref(&self) -> &PhoneNumberRef {
        self.borrow()
    }
}

impl Deref for PhoneNumber {
    // Not sure if this is useful or if it should just deref into a &[u8]
    type Target = PhoneNumberRef;

    fn deref(&self) -> &Self::Target {
        self.borrow()
    }
}

impl PartialEq<PhoneNumberRef> for PhoneNumber {
    fn eq(&self, other: &PhoneNumberRef) -> bool {
        self.0.eq(&other.0)
    }
}

#[derive(Debug, PartialEq, Eq, Hash)]
#[repr(transparent)]
pub struct PhoneNumberRef(pub(crate) [u8]);

impl PhoneNumberRef {
    /// Phone numbers must be 11 digits and start with 0
    pub fn new(value: &[u8]) -> ValidationResult<&Self> {
        if value.len() != 11 {
            Err(PhoneNumberError::InvalidLength)?;
        }

        if value[0] != b'0' {
            Err(PhoneNumberError::InvalidFirstCharacter)?;
        }

        for (i, b) in value.iter().enumerate() {
            if !b.is_ascii_digit() {
                Err(PhoneNumberError::InvalidCharacter(i, *b))?;
            }
        }

        // Safety:
        // PhoneNumberRef is a wrapper of str, therefore converting a &str to a &PhoneNumberRef
        // is safe.
        Ok(unsafe { &*(value as *const _ as *const Self) })
    }

    pub fn as_str(&self) -> &str {
        // Safety:
        // The validation in the constructor makes sure that all the bytes are ascii.
        unsafe { str::from_utf8_unchecked(&self.0) }
    }
}

impl ToOwned for PhoneNumberRef {
    type Owned = PhoneNumber;

    fn to_owned(&self) -> Self::Owned {
        // Unwrap:
        // This is fine since we checked the length when we constructed the PhoneNumberRef
        PhoneNumber(self.0.try_into().unwrap())
    }
}

impl PartialEq<PhoneNumber> for PhoneNumberRef {
    fn eq(&self, other: &PhoneNumber) -> bool {
        self.0.eq(&other.0)
    }
}
